/*
 * Copyright (C) 2007 Moritz Orbach <web@apfelboymchen.homeunix.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * disclaimer:
 * I guess this isn't the best code. I just wrote the functions without 
 * thinking much about them after it started to work. So there may be the the 
 * one or other segfault lurking.
 *
 */
#include <stdlib.h>
#include <math.h>
#include "SDL/SDL.h"
#include "SDL/SDL_ttf.h" 

#include "state.h"

/*
 * http://www.libsdl.org/cgi/docwiki.cgi/SDL_20API
 * http://www.libsdl.org/cgi/docwiki.cgi/SDL_5fttf
 * http://www.libsdl.org/cgi/docwiki.cgi/Examples
 * http://www.linuxdevcenter.com/pub/a/linux/2003/05/15/sdl_anim.html
 * http://gpwiki.org/index.php/SDL_ttf:Tutorials:Basic_Font_Rendering
 * http://www.gamedev.net/reference/articles/article1960.asp
 * http://www.gamedev.net/reference/articles/article1953.asp
 */


unsigned int fps;


SDL_Surface *screen = NULL;
TTF_Font *font = NULL;
t_state state;

/*
 * stolen from ... somewhere
 * putpixel (x=321, y=863, color=16536660)
 *           x=606, y=863, color=16536660
 * x: 276
 * y: 534
 *
 */
void putpixel(int x, int y, int color)
{ 
	unsigned int *ptr = (unsigned int*)screen->pixels;
	/* printf("x: %d\ny: %d\n", x, y); */

	/* printf("%d:%d\n", x, y); */

	int lineoffset = y * (screen->pitch / 4);
	ptr[lineoffset + x] = color;

	if (y > 1) {
		lineoffset = (y - 1) * (screen->pitch / 4);
		ptr[lineoffset + x] = color;
	}

	/* first pixel is 0 --> 0 to HEIGHT - 1 */
	if (y < HEIGHT - 1) {
		lineoffset = (y + 1) * (screen->pitch / 4);
		ptr[lineoffset + x] = color;
	}

	if (x > 1) {
		lineoffset = y * (screen->pitch / 4);
		ptr[lineoffset + x - 1] = color;
	}

	if (x < WIDTH - 2) {
		lineoffset = y * (screen->pitch / 4);
		ptr[lineoffset + x + 1] = color;
	}
}

void ttf_printline(char *s, int lineno)
{
	SDL_Surface *text;
	SDL_Color color = { 67, 154, 175, 64 };

	int lineskip = TTF_FontLineSkip(font);

	text = TTF_RenderText_Blended(font, s, color);

	SDL_Rect rcDest = { lineskip, lineskip * lineno, lineskip, lineskip };
	SDL_BlitSurface(text, NULL, screen, &rcDest);
	SDL_FreeSurface(text);
	/* printf("=%s=\n", s); */
}

/* this sucks, but it was quick */
void ttf_printlines(char *s, int startline)
{
	int spos = 0;
	int dpos;
	char line[STRLEN];
	int lineno = startline;

	while(s[spos] != '\0') {
		dpos = 0;
		while(s[spos] != '\0' && dpos < STRLEN && s[spos] != '\n') {
			line[dpos] = s[spos];
			dpos++;
			spos++;
		}
		/* \0 || dpos == STRLEN || \n */
		line[dpos] = '\0';
		if (s[spos] != '\0')
			spos++;
		lineno++;
		/* printf("descht >%s<\n", line); */
		ttf_printline(line, lineno);
	}
}


void showconfig()
{
	char s[STRLEN];
	snprintf(s, STRLEN,
			"key          variable     value"
			"             FPS: %d\n"
			"q/a          y_amplitude: %d\n"
			"w/s          x_amplitude: %d\n"
			"e/d          y_frequency: %.3f\n"
			"r/f          x_frequency: %.3f\n"
			"t/g          y_speed    : %.3f\n"
			"z/h          x_speed    : %.3f\n"
			"x/y          slope      : %d\n"
			"c/v          spacing    : %d\n"
			"\n"
			"F2           more keys",
			fps,
			state.config->y_amplitude,
			state.config->x_amplitude,
			state.config->y_frequency,
			state.config->x_frequency,
			state.config->y_speed,
			state.config->x_speed,
			state.config->slope,
			state.config->spacing);


	ttf_printlines(s, 1);
}

void morekeys()
{
	char *s = 
		"SPACE        toggle speed\n"
		"k/m          height\n"
		",/.          width\n"
		"arrows       move schwabbel\n"
		"INS/DEL      change color (red)\n"
		"HOME/END     change color (green)\n"
		"PGUP/PGDN    change color (blue)\n"
		"TAB          toggle fullscreen\n"
		"[0-9]        load slot\n"
		"shift-[0-9]  save to slot\n"
		"K_Enter      save to current slot\n"
		"TAB          toggle fullscreen\n"
		"F1           toggle variables";

	ttf_printlines(s, 12);

}

/* ----------- */

/*
 * eine sinus-welle mit startwert sin(wellex)
 */
void draw_sinus(int offset_x, int offset_y, int xlen, double wellex, int y_depth)
{
	int pixelpos_x;
	int pixelpos_y;

	/* y * 2: links und rechts frei lassen */
	for (pixelpos_x = offset_x; pixelpos_x < offset_x + xlen; pixelpos_x += state.config->spacing) {
		pixelpos_y = (int)(sin(wellex) * state.config->x_amplitude) + offset_y;

		if (pixelpos_x < WIDTH && pixelpos_x >= 0 && pixelpos_y < HEIGHT && pixelpos_y >= 0)
			putpixel(pixelpos_x, pixelpos_y, state.config->color - ( (pixelpos_y - offset_y - state.config->spacing + y_depth) * 0x010101));
			/* 010101 == change color value, not color */
			/* putpixel(pixelpos_x, pixelpos_y, 0xFC5454 + (offset_y - pixelpos_y) * 0x30301); */
			/* gibt 41996B, von 110 198 122  auf  2 90 86 */
			/* putpixel(pixelpos_x, pixelpos_y, 0xFC5454 + wellex); */
		/*
		else
			printf("%d > %d || %d > %d\n", pixelpos_x, WIDTH, pixelpos_y, HEIGHT);
		 */
		/* printf("%d\n", (pixelpos_y - offset_y)); */
		wellex += state.config->x_frequency;
	}
}

void draw()
{   
	static double startval_x = 0;
	static double startval_y = 0;
	static double startval_flagmove = 0;
	double wellex_sin = startval_x;
	double welley_sin = startval_y;
	/* increment for next call */
	startval_x += state.config->x_speed;
	startval_y += state.config->y_speed;

	/* Lock surface if needed*/
	if (SDL_MUSTLOCK(screen) && SDL_LockSurface(screen) < 0) {
		printf("warnix\n");
		return;
	}

	SDL_FillRect(screen, 0, 0);


	/* ---------- */

	int flag_pixel_x = state.config->border_left;
	int flag_pixel_y;

	/* fahnen-offset */

	/* durchlaeufe unabhaengig von restlichem platz / immer feste zeilenzahl */
	int flag_lines = state.config->flag_height / state.config->spacing;

	/* ugh! */
	int i;
	for (i = 1; i < flag_lines; i++) {
		/* neue zeile */
		flag_pixel_y = state.config->border_top + i * (state.config->spacing);
		welley_sin += state.config->y_frequency;
		int pixelpos_y = sin(welley_sin) * state.config->y_amplitude + flag_pixel_y;
		draw_sinus(flag_pixel_x,
			   pixelpos_y,
			   state.config->flag_width, wellex_sin,
			   (sin(welley_sin) * state.config->y_amplitude));
		/* wie weit es pro zeile nach rechts geht */
		flag_pixel_x += state.config->slope;

	}
	/* !hgu */

	/* reset startval_x. 4PI and 0 are equal */
	if (startval_x > M_PI * 4) {
		startval_x = 0;
	}
	if (startval_y > M_PI * 4) {
		startval_y = 0;
	}
	if (startval_flagmove > M_PI * 4) {
		startval_flagmove = 0;
	}
	/*
	*/

	if (state.help) {
		showconfig();
	}
	if (state.morehelp) {
		morekeys();
	}
	/* ---------- */


	/* Unlock if needed*/
	if (SDL_MUSTLOCK(screen))
		SDL_UnlockSurface(screen);

	/*
	 * Tell SDL to update the whole screen
	 * this would be the same: SDL_UpdateRect(screen, 0, 0, WIDTH, HEIGHT);
	 */
	SDL_Flip(screen);

}

/* ----------- */


void check_surface()
{
	if ( screen == NULL ) {
		fprintf(stderr, "Unable to set video-mode: %s\n", SDL_GetError());
		exit(1);
	}
}

/*
 * The option SDL_ANYFORMAT tells SDL to leave the number of bits per pixel the 
 * same as the current screen settings. 
 */
/* using a local variable doesn't work?! */
void fullscreen(SDL_Surface **screen)
{
	if (*screen != NULL) {
		/* free old */
		SDL_FreeSurface(*screen);
		printf("going fullscreen\n");
	}

	*screen = SDL_SetVideoMode(WIDTH, HEIGHT, 32, SDL_ANYFORMAT|SDL_FULLSCREEN);
	check_surface();
}

/* using a local variable doesn't work?! */
void window(SDL_Surface **screen)
{
	/* bei fehlschlag im fenster bleiben lohnt eigentlich nicht */
	if (*screen != NULL) {
		/* free old */
		SDL_FreeSurface(*screen);
		printf("leaving fullscreen\n");
	}

	*screen = SDL_SetVideoMode(WIDTH, HEIGHT, 32, SDL_ANYFORMAT|SDL_HWSURFACE);
	check_surface();
}


/* using a local variable doesn't work?! */
void loadfont(TTF_Font **font)
{

	/* font-stuff */
	char *fontfile = FONTPATH;

	if (TTF_Init() == -1) {
		printf("Unable to initialize SDL_ttf: %s\n", TTF_GetError());
		exit(2);
	}

	*font = TTF_OpenFont(fontfile, 11);
	if (*font == NULL) {
		/* printf("Unable to load font: %s %s\n", fontfile, TTF_GetError); */
		printf("gehtnet\n");
		exit(3);
	}
}


/* -------- */

void quit()
{
	TTF_Quit();
	SDL_Quit();
}


int main(int argc, char *argv[])
{


	/* Initialize SDL's subsystems - in this case, only video.*/
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
		exit(1);
	}

	atexit(quit);
	loadfont(&font);
	window(&screen);


	/* delay, interval */
	SDL_EnableKeyRepeat(50, 20);
	SDL_ShowCursor(SDL_DISABLE);
	SDL_EnableUNICODE(1);

	/* init(); */
	load_config();
	/* todo: 0 default */
	if (!loadslot(&state.config, 1)) {
		printf("couldn't load config!\n");
		exit(1);
	}
	/*
	if (!state.config) {
		printf("haaaaaeeee?!\n");
		exit(1);
	}
	*/

	/* Main loop */

	typedef struct {
		int stopped;
		double x;
		double y;
	} t_speed;

	t_speed speed = { 0, 0, 0 };
	int delay = 10;
	while (1) {

		unsigned int before = SDL_GetTicks();
		draw();
		/* void SDL_Delay(Uint32 ms); */
		SDL_Delay(delay);
		unsigned int duration = SDL_GetTicks() - before;
		/* printf("%d = 1000 / (%d + %d)\n", fps, duration, delay); */

		/*
		 * 1000/WANTED_FRAMERATE        = should-duration
		 * should-duration              = duration + delay
		 * 1000/WANTED_FRAMERATE        = duration + delay
		 *
		 * Uint32 SDL_GetTicks(void)
		 *   Gets the number of milliseconds since SDL library initialization.
		 *   Returns the number of milliseconds since SDL library initialization. 
		 *   Note that this value wraps around if the program runs for more than 
		 *   ~49.7 days.
		 * http://www.libsdl.org/cgi/docwiki.cgi/SDL_5fGetTicks
		 * http://www.libsdl.org/cgi/docwiki.cgi/SDL_20Average_20FPS_20Measurement
		 */
		delay = 1000/WANTED_FRAMERATE - duration;
		if (delay < 1)
			delay = 1;
		fps = 1000 / (duration + delay);
		/*
		printf("delay   : %d\n", delay);
		printf("duration: %d\n\n", duration);
		*/

		/* Poll for events, and handle the ones we care about.*/
		SDL_Event event;
		while (SDL_PollEvent(&event)) {
			int shift = SDL_GetModState() & KMOD_SHIFT;
			switch (event.type) {

				case SDL_KEYDOWN:
					switch (event.key.keysym.sym) {
						case SDLK_ESCAPE:
							return 0;


							/* amplitude */
						case SDLK_w:
							state.config->x_amplitude += 1;
							break;
						case SDLK_s:
							if (state.config->x_amplitude > 0)
								state.config->x_amplitude -= 1;
							break;

						case SDLK_q:
							state.config->y_amplitude += 1;
							break;
						case SDLK_a:
							if (state.config->y_amplitude > 0)
								state.config->y_amplitude -= 1;
							break;


							/* frequency */
						case SDLK_e:
							state.config->y_frequency += 0.01;
							break;
						case SDLK_d:
							state.config->y_frequency -= 0.01;
							break;

						case SDLK_r:
							state.config->x_frequency += 0.01;
							break;
						case SDLK_f:
							state.config->x_frequency -= 0.01;
							break;


						case SDLK_y:
							state.config->slope -= 1;
							break;
						case SDLK_x:
							state.config->slope += 1;
							break;

						case SDLK_v:
							if (state.config->spacing > 1)
								state.config->spacing -= 1;
							break;
						case SDLK_c:
							state.config->spacing += 1;
							break;

							/* colors */
						case SDLK_INSERT:
							state.config->color += 0x010000;
							break;
						case SDLK_DELETE:
							state.config->color -= 0x010000;
							break;
						case SDLK_HOME:
							state.config->color += 0x000100;
							break;
						case SDLK_END:
							state.config->color -= 0x000100;
							break;
						case SDLK_PAGEUP:
							state.config->color += 0x000001;
							break;
						case SDLK_PAGEDOWN:
							state.config->color -= 0x000001;
							break;

							/* speed */
						case SDLK_t:
							state.config->y_speed += 0.01;
							break;

						case SDLK_g:
							state.config->y_speed -= 0.01;
							break;
						case SDLK_z:
							state.config->x_speed += 0.01;
							break;

						case SDLK_h:
							state.config->x_speed -= 0.01;
							break;


						case SDLK_SPACE:
							if (speed.stopped) {
								speed.stopped = 0;
								state.config->x_speed = speed.x;
								state.config->y_speed = speed.y;
							} else {
								speed.stopped = 1;
								speed.x = state.config->x_speed;
								speed.y = state.config->y_speed;
								state.config->x_speed = 0;
								state.config->y_speed = 0;
							}
							break;

							/* position */
						case SDLK_UP:
							state.config->border_top -= state.config->spacing;
							break;
						case SDLK_DOWN:
							state.config->border_top += state.config->spacing;
							break;
						case SDLK_LEFT:
							state.config->border_left -= state.config->spacing;
							break;
						case SDLK_RIGHT:
							state.config->border_left += state.config->spacing;
							break;

							/* size */
						case SDLK_PERIOD:
							state.config->flag_width += state.config->spacing;
							break;
						case SDLK_COMMA:
							if ((state.config->flag_width - state.config->spacing) / state.config->spacing > 1)
								state.config->flag_width -= state.config->spacing;
							break;
						case SDLK_k:
							if ((state.config->flag_height - state.config->spacing) / state.config->spacing > 1)
								state.config->flag_height -= state.config->spacing;
							break;
						case SDLK_m:
							state.config->flag_height += state.config->spacing;
							break;


						case SDLK_TAB:
							if (state.fullscreen) {
								window(&screen);
							} else {
								fullscreen(&screen);
							}
							state.fullscreen = !state.fullscreen;
							break;

						case SDLK_F1:
							state.help = !state.help;
							break;
						case SDLK_F2:
							state.morehelp = !state.morehelp;
							break;

							/* config */
						case SDLK_1:
							if (!shift) {
								loadslot(&state.config, 1);
								state.currentslot = 1;
							} else {
								save(1);
							}
							break;
						case SDLK_2:
							if (!shift) {
								loadslot(&state.config, 2);
								state.currentslot = 2;
							} else {
								save(1);
							}
							break;
						case SDLK_3:
							if (!shift) {
								loadslot(&state.config, 3);
								state.currentslot = 3;
							} else {
								save(3);
							}
							break;
						case SDLK_4:
							if (!shift) {
								loadslot(&state.config, 4);
								state.currentslot = 4;
							} else {
								save(4);
							}
							break;
						case SDLK_5:
							if (!shift) {
								loadslot(&state.config, 5);
								state.currentslot = 5;
							} else {
								save(5);
							}
							break;
						case SDLK_6:
							if (!shift) {
								loadslot(&state.config, 6);
								state.currentslot = 6;
							} else {
								save(6);
							}
							break;
						case SDLK_7:
							if (!shift) {
								loadslot(&state.config, 7);
								state.currentslot = 7;
							} else {
								save(7);
							}
							break;
						case SDLK_8:
							if (!shift) {
								loadslot(&state.config, 8);
								state.currentslot = 8;
							} else {
								save(8);
							}
							break;
						case SDLK_9:
							if (!shift) {
								loadslot(&state.config, 9);
								state.currentslot = 9;
							} else {
								save(9);
							}
							break;
						case SDLK_0:
							if (!shift) {
								loadslot(&state.config, 0);
								state.currentslot = 0;
							} else {
								save(0);
							}
							break;
						case SDLK_KP_ENTER:
							save(state.currentslot);
							break;
						case SDLK_CARET:
							printf("reset\n");
							load_config();
							break;

						default:
							/* printf("unknown key: %d\n", event.key.keysym.sym); */
							break;
					}

					/*
					 * astersik conflicts for non-german keyboard-layouts
					printf("%d\n", event.key.keysym.unicode);
					switch (event.key.keysym.unicode) {
						case SDLK_ASTERISK:
							save(state.currentslot);
							break;
					}
					*/

			} /* even type */
		}
	}
	return 0;
}


